package edu.umd.dawn.common.interceptor;

import edu.umd.dawn.common.logging.RequestLog;
import edu.umd.dawn.common.services.UserAgentService;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import lombok.extern.log4j.Log4j2;
import org.apache.logging.log4j.Level;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.HandlerMapping;

@Log4j2
@Component
public class RequestInterceptor implements HandlerInterceptor {

    private String TRACE_HEADER = "X-Request-Id";
    
    @Value("${config.serviceName:dawn-service}")
    private String serviceName;

    @Autowired
    private UserAgentService userAgentService;

    @Override
    public boolean preHandle(HttpServletRequest request, HttpServletResponse response, Object handler)
            throws Exception {
        String traceID = UUID.randomUUID().toString();
        if (request.getHeader(TRACE_HEADER) != null
                && !request.getHeader(TRACE_HEADER).isEmpty()) {
            traceID = request.getHeader(TRACE_HEADER);
        }
        MDC.put("requestId", traceID);
        MDC.put("serviceName", serviceName);

        response.setHeader(TRACE_HEADER, traceID);

        long startTime = System.currentTimeMillis();
        request.setAttribute("startTime", startTime);
        return true;
    }

    private Map<String, String> getRequestHeaders(HttpServletRequest request) {
        Enumeration<String> headerNames = request.getHeaderNames();
        Map<String, String> headers = new HashMap<>();

        if (headerNames != null) {
            while (headerNames.hasMoreElements()) {
                String ele = headerNames.nextElement();
                headers.put(ele, request.getHeader(ele));
            }
        }
        return headers;
    }

    @SuppressWarnings("unchecked")
    private Map<String, String> getParemeters(HttpServletRequest request) {
        Enumeration<String> parameterNames = request.getParameterNames();
        Map<String, String> parameters = new HashMap<>();

        if (parameterNames != null) {
            while (parameterNames.hasMoreElements()) {
                String ele = parameterNames.nextElement();
                parameters.put(ele, request.getParameter(ele));
            }
        }

        Object p = request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);

        if (p != null) {
            Map<String, String> pathParams = (Map<String, String>) p;
            pathParams.forEach((k, v) -> parameters.merge(k, v, String::concat));
        }

        return parameters;
    }

    @Override
    public void afterCompletion(
            HttpServletRequest request, HttpServletResponse response, Object handler, Exception ex) {
        long startTime = (Long) request.getAttribute("startTime");
        long endTime = System.currentTimeMillis();
        long executeTime = endTime - startTime;

        String fullPath = request.getRequestURI();
        String servletPattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);
        String servletPath = request.getServletPath();

        String path = fullPath.replace(servletPath, servletPattern);

        Map<String, String> headers = getRequestHeaders(request);

        RequestLog requestLog = RequestLog.builder()
                .path(path)
                .method(request.getMethod())
                .statusCode(response.getStatus())
                .headers(headers)
                .parameters(getParemeters(request))
                .duration(executeTime)
                .userAgent(userAgentService.parse(headers))
                .build();

        log.log(Level.forName("REQUEST", 10), requestLog);
    }
}
